use std::process::ExitCode;

fn main() -> ExitCode {
    if std::env::args().len() != 2 {
        eprintln!("Must provide one argument for the test input");
        return ExitCode::FAILURE;
    }
    let argument = std::env::args().into_iter().skip(1).next().unwrap();
    match argument.as_str() {
        "-h" | "--help" => println!("{}", usage()),
        input => {
            match input
                .split_whitespace()
                .map(|to_parse| to_parse.parse::<u32>().map(|parsed| parsed as i32))
                .collect::<Result<Vec<i32>, _>>()
            {
                Ok(seq) => {
                    let actual = jump_game::Solution::can_jump(seq);
                    println!("{}", actual);
                }
                Err(_error) => {
                    eprintln!("Input is not in a valid format");
                    return ExitCode::FAILURE;
                }
            }
        }
    }

    return ExitCode::SUCCESS;
}

fn usage() -> &'static str {
    "jump_game [sequence as input]\n\
    Syntax for sequence (\\d+\\s)\n\
    Example: 2 4"
}
